import pyperclip3 as pc
from os import remove
from os.path import getctime, realpath, exists
from datetime import datetime
from itertools import groupby
from sys import executable
from sys import exit as sysexit

from cat_win.util.ArgConstants import *
import cat_win.util.ArgParser as ArgParser
import cat_win.util.checksum as checksum
import cat_win.util.Converter as Converter
import cat_win.util.Holder as Holder
import cat_win.util.StdInHelper as StdInHelper

from cat_win import __version__, __author__, __sysversion__

holder = Holder.Holder()
converter = Converter.Converter()

def _showHelp():
    print("Usage: cat [FILE]... [OPTION]...")
    print("Concatenate FILE(s) to standard output.")
    print()
    for x in ALL_ARGS:
        print("%-25s" % str("\t" + x.shortForm + ", " + x.longForm), end=x.help + "\n")
    print()
    print("%-25s" % str("\t'enc=X':"), end="set file encoding to X.\n")
    print()
    print("%-25s" % str("\t'[a;b]':"), end="replace a with b in every line.\n")
    print("%-25s" % str("\t'[a:b]':"), end="python-like string manipulation syntax.\n")
    print()
    print("Examples:")
    print("%-25s" % str("\tcat f g -r"), end="Output g's contents in reverse order, then f's content in reverse order\n")
    print("%-25s" % str("\tcat f g -ne"), end="Output f's, then g's content, while numerating and showing the end of lines.\n")
    sysexit(0)

def _showVersion():
    print()
    print("------------------------------------------------------------")
    print("Cat", __version__)
    print("------------------------------------------------------------")
    print()
    print("Python: \t", __sysversion__) #sys.version
    print("Build time: \t " + str(datetime.fromtimestamp(getctime(realpath(executable)))) + " CET")
    print("Author: \t", __author__)
    sysexit(0)

def _showDebug(args, known_files, unknown_files):
    print("Debug Information:")
    print("args: ", end="")
    print(args)
    print("known_files:", end="")
    print(known_files)
    print("unknown_files:", end="")
    print(unknown_files)

def _getLinePrefix(index, line_num):
    line_prefix = str(line_num) + ") "
    for i in range(len(str(line_num)), holder.fileLineMaxLength):
        line_prefix += " "
    
    file_prefix = ""
    if len(holder.files) > 1:
        file_prefix += str(index)
        for i in range(len(str(index)), holder.fileMaxLength):
            file_prefix += " "
        file_prefix += "."
        
    return file_prefix + line_prefix
    
def printFile(fileIndex = 1):
    show_bytecode = False
    content = []
    try:
        with open(holder.files[fileIndex-1], 'r', encoding=ArgParser.FILE_ENCODING) as f:
            content = f.read().splitlines()
    except:
        print("Failed to open:", holder.files[fileIndex-1])
        print("Do you want to open the file as a binary without parameters? [Y]")
        inp = input()
        if not 'y' in inp and not 'Y' in inp: return
        try:
            with open(holder.files[fileIndex-1], 'rb') as f:
                content = f.read().splitlines()
            show_bytecode = True
        except:
            print("Operation failed! Try using the enc=X parameter.")
            return
    fLength = len(content)
    if not show_bytecode:
        for i, arg in enumerate(holder.args_id):
            if arg == ARGS_NUMBER:
                content = [_getLinePrefix(fileIndex, holder.fileCount-i if holder.reversed else holder.fileCount+i+1) + c for i, c in enumerate(content)]
                holder.fileCount += (-fLength if holder.reversed else fLength)
            if arg == ARGS_ENDS:
                content = [c + "$" for c in content]
            if arg == ARGS_TABS:
                content = [c.replace("\t", "^I") for c in content]
            if arg == ARGS_SQUEEZE:
                content = [g[0] for g in groupby(content)]
            if arg == ARGS_REVERSE:
                content.reverse()
            if arg == ARGS_BLANK:
                content = [c for c in content if c]
            if arg == ARGS_DEC:
                if holder.args[i][1] == "-dec":
                    content = [converter._fromDEC(int(c), True) for c in content if converter.is_decimal(c)]
                else:
                    content = [converter._fromDEC(int(c)) for c in content if converter.is_decimal(c)]
            if arg == ARGS_HEX:
                if holder.args[i][1] == "-hex":
                    content = [converter._fromHEX(c, True) for c in content if converter.is_hex(c)]
                else:
                    content = [converter._fromHEX(c) for c in content if converter.is_hex(c)]
            if arg == ARGS_BIN:
                if holder.args[i][1] == "-bin":
                    content = [converter._fromBIN(c, True) for c in content if converter.is_bin(c)]
                else:
                    content = [converter._fromBIN(c) for c in content if converter.is_bin(c)]
            if arg == HIGHEST_ARG_ID+1:
                try:
                    content = [eval(repr(c) + holder.args[i][1]) for c in content]
                except:
                    print("Error at operation: ", holder.args[i][1])
                    return
            if arg == HIGHEST_ARG_ID+2:
                replace_values = holder.args[i][1][1:-1].split(";")
                content = [c.replace(replace_values[0], replace_values[1]) for c in content]
    
    print(*content, sep="\n")
    
    if not show_bytecode:
        if ARGS_CLIP in holder.args_id:
            holder.clipBoard += "\n".join(content)

def printFiles():
    start = len(holder.files)-1 if holder.reversed else 0
    end = -1 if holder.reversed else len(holder.files)
    if ARGS_CHECKSUM in holder.args_id:
        for file in holder.files:
            print("Checksum of '" + file + "':")
            print("type", type(file))
            print(checksum.getChecksumFromFile(file))
    else:
        for i in range(start, end, -1 if holder.reversed else 1):
            printFile(i+1)
        if ARGS_COUNT in holder.args_id:
            print()
            print("Lines: " + str(holder.lineSum))
        if ARGS_FILES in holder.args_id:
            print()
            print("applied FILE(s):", end="")
            print("", *holder.files, sep="\n\t")
        if ARGS_CLIP in holder.args_id:
            pc.copy(holder.clipBoard)
            
def main():
    piped_input = temp_file = ""
    
    #read parameter-args
    args, known_files, unknown_files = ArgParser.getArguments()
    holder.setArgs(args)
    
    #check for special cases
    if (len(known_files) == 0 and len(unknown_files) == 0 and len(holder.args) == 0) or ARGS_HELP in holder.args_id: _showHelp()
    if ARGS_VERSION in holder.args_id: _showVersion()
    if ARGS_DEBUG in holder.args_id: _showDebug(args, known_files, unknown_files)
    if ARGS_INTERACTIVE in holder.args_id:
        piped_input = StdInHelper.getStdInContent()
        temp_file = StdInHelper.writeTemp(piped_input)
        known_files.append(temp_file)
        StdInHelper.writeFiles(unknown_files, piped_input)
    else:
        StdInHelper.readWriteFilesFromStdIn(unknown_files)
    
    #fill holder object with neccessary values
    holder.setFiles([*known_files, *unknown_files])
    holder.generateValues()
    
    #print the cat-output
    printFiles()
    
    #clean-up
    if exists(temp_file):
        remove(temp_file)
    
if __name__ == "__main__":
    main()
#pyinstaller cat.py --onefile --clean --dist ../bin


class ArgConstant():
    def __init__(self, shortForm, longForm, help, id):
        self.shortForm = shortForm
        self.longForm = longForm
        self.help = help
        self.id = id

ARGS_HELP = 0
ARGS_NUMBER = 1
ARGS_ENDS = 2
ARGS_TABS = 3
ARGS_SQUEEZE = 4
ARGS_REVERSE = 5
ARGS_COUNT = 6
ARGS_BLANK = 7
ARGS_FILES = 9
ARGS_INTERACTIVE = 10
ARGS_CLIP = 11
ARGS_CHECKSUM = 12
ARGS_DEC = 13
ARGS_HEX = 14
ARGS_BIN = 15
ARGS_VERSION = 16
ARGS_DEBUG = 17

ALL_ARGS = [[["-h", "--help"], "show this help message and exit", ARGS_HELP],
            [["-n", "--number"], "number all output lines", ARGS_NUMBER],
            [["-e", "--ends"], "display $ at end of each line", ARGS_ENDS],
            [["-t", "--tabs"], "display TAB characters as ^I", ARGS_TABS],
            [["-s", "--squeeze"], "suppress repeated output lines", ARGS_SQUEEZE],
            [["-r", "--reverse"], "reverse output", ARGS_REVERSE],
            [["-c", "--count"], "show sum of lines", ARGS_COUNT],
            [["-b", "--blank"], "hide empty lines", ARGS_BLANK],
            [["-f", "--files"], "list applied files", ARGS_FILES],
            [["-i", "--interactive"], "use stdin", ARGS_INTERACTIVE],
            [["-l", "--clip"], "copy output to clipboard", ARGS_CLIP],
            [["-m", "--checksum"], "show the checksums of all files", ARGS_CHECKSUM],
            [["-dec", "--dec"], "convert decimal number to hexadecimal and binary", ARGS_DEC],
            [["-hex", "--hex"], "convert hexadecimal number to decimal and binary", ARGS_HEX],
            [["-bin", "--bin"], "convert binary number to decimal and hexadecimal", ARGS_BIN],
            [["-v", "--version"], "output version information and exit", ARGS_VERSION],
            [["-d", "--debug"], "show debug information", ARGS_DEBUG]]

ALL_ARGS = [ArgConstant(x[0][0], x[0][1], x[1], x[2]) for x in ALL_ARGS]
HIGHEST_ARG_ID = max([m.id for m in ALL_ARGS])


from cat_win.util.ArgConstants import *

class Holder():
    files = []
    args = []
    args_id = []
    reversed = False
    lineSum = 0
    fileCount = 0
    fileLineMaxLength = 0
    fileMaxLength = 0
    clipBoard = ""
    
    def setFiles(self, files: list):
        self.files = files
    
    def setArgs(self, args: list):
        self.args = args
        self.args_id = [x[0] for x in self.args]
        self.reversed = ARGS_REVERSE in self.args_id
    
    def __count_generator__(self, reader):
        b = reader(1024 * 1024)
        while b:
            yield b
            b = reader(1024 * 1024)
        
    def __getFileLinesSum__(self, file):
        with open(file, 'rb') as fp:
            c_generator = self.__count_generator__(fp.raw.read)
            count = sum(buffer.count(b'\n') for buffer in c_generator)
            return count + 1
    
    def __calcFilesLineSum__(self):
        self.lineSum = sum([self.__getFileLinesSum__(file) for file in self.files])
        self.fileCount = self.lineSum if self.reversed else 0
        
    def __calcFileLineMaxLength__(self):
        self.fileLineMaxLength = len(str(self.fileCount)) if self.reversed else len(str(self.lineSum))

    def __calcFileMaxLength__(self):
        self.fileMaxLength = len(str(len(self.files)))
        
    def generateValues(self):
        self.__calcFilesLineSum__()
        self.__calcFileLineMaxLength__()
        self.__calcFileMaxLength__()


import hashlib
from zlib import crc32 as crc32_hash

def getChecksumFromFile(file: str):
    """
    Takes a filepath of type String and
    returns a String representation of the
    CRC32, MD5, SHA1, SHA256, SHA512
    hashes corresponding to the given file.
    """
    BUF_SIZE = 65536 #64kb
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()
    sha512 = hashlib.sha512()
    crc32 = 0
    
    with open(file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
            sha1.update(data)
            sha256.update(data)
            sha512.update(data)
            crc32 = crc32_hash(data, crc32) 
    
    crc32 = "%08X" % (crc32 & 0xFFFFFFFF)
    
    checksum  = "%-10s" % "\tCRC23:"  + str(crc32) + "\n"
    checksum += "%-10s" % "\tMD5:"    + str(md5.hexdigest()) + "\n"
    checksum += "%-10s" % "\tSHA1:"   + str(sha1.hexdigest()) + "\n"
    checksum += "%-10s" % "\tSHA256:" + str(sha256.hexdigest()) + "\n"
    checksum += "%-10s" % "\tSHA512:" + str(sha512.hexdigest()) + "\n"
    return checksum


from sys import argv
from glob import iglob
from re import match
from os.path import isfile, realpath, isdir
from cat_win.util.ArgConstants import *

FILE_ENCODING = None

def __addArgument__(args, known_files, unknown_files, param):
    if match("\Aenc\=.+\Z", param):
        global FILE_ENCODING
        FILE_ENCODING = param[4:]
        return
    elif match("\A\[.*\:.*\]\Z", param):
        args.append([HIGHEST_ARG_ID+2, param])
        return
    elif match("\A\[.+\;.+\]\Z", param):
        args.append([HIGHEST_ARG_ID+3, param])
        return
    
    for x in ALL_ARGS:
        if x.shortForm == param or x.longForm == param:
            args.append([x.id, param])
            return
        
    possible_path = realpath(param)
    if match("\*", param):
        for filename in iglob("./" + param, recursive=True):
            if isdir(filename):
                continue
            known_files.append(realpath(filename))
        return       
    elif isdir(possible_path):
        for filename in iglob(possible_path + '**/**', recursive=True):
            if isdir(filename):
                continue
            known_files.append(realpath(filename))
        return
    elif isfile(possible_path):
        known_files.append(realpath(possible_path))
        return
    elif param[0] == "-" and len(param) > 2:
        for i in range(1, len(param)):
            __addArgument__(args, known_files, unknown_files, "-" + param[i])
    elif match("\A[^-]+\Z", param):
        unknown_files.append(realpath(param))

def getArguments():
    inputArgs = argv
    args = []
    known_files = []
    unknown_files = []
    
    for i in range(1, len(inputArgs)): __addArgument__(args, known_files, unknown_files, inputArgs[i])
        
    return (args, known_files, unknown_files)


from sys import stdin
from tempfile import NamedTemporaryFile

def writeTemp(content: str):
    """
    Writes content into a generated temp-file and
    returns the path in type String.
    """
    tmp_file = NamedTemporaryFile(delete=False).name
    with open(tmp_file, 'w') as f:
        f.write(content)
    return tmp_file

def getStdInContent():
    """
    returns a String delivered by the standard input.
    """
    input = ""
    for line in stdin:
        input += line
    return input
    
def writeFiles(file_list: list, content: str):
    """
    Simply writes the content into every
    file in the given list if there is a
    valid content.
    """
    if content == "": file_list.clear()
    for file in file_list:
        with open(file, 'w') as f:
            f.write(content)
    
    
def readWriteFilesFromStdIn(file_list: list):
    """
    Takes a list of files, waits for a String from
    the standard input and writes it into every file.
    """
    if len(file_list) == 0: return
    
    print("The given FILE(s)", end="")
    print("", *file_list, sep="\n\t")
    print("do/does not exist. Write the FILE(s) and finish with the '^Z'-suffix ((Ctrl + Z) + Enter):")
    
    input =  getStdInContent()
    input = input.rstrip()
    if len(input) > 0 and ord(input[-1:]) == 26: input = input[:-1]
    
    writeFiles(file_list, input)


import string

class Converter():
    def is_decimal(self, v: str):
        """
        returns True if given String is a Decimal number.
        return False if it is not.
        """
        return v.isdecimal()

    def is_hex(self, v: str):
        """
        returns True if a String is a Hexadecimal number.
        returns False if it is not.
        """
        hex_digits = set(string.hexdigits)
        if v[:2] == "0x": v = v[2:]
        return all(c in hex_digits for c in v)

    def is_bin(self, v: str):
        """
        returns True if a String is a Binary number.
        returns False if it is not.
        """
        v_set = set(v)
        return v_set == {'0', '1'} or v_set == {'0'} or v_set == {'1'}

    def __dec_to_hex__(self, value: int, leading: bool = False):
        return "{0:#x}".format(value) if leading else "{0:x}".format(value)

    def __dec_to_bin__(self, value: int, leading: bool = False):
        return "{0:#b}".format(value) if leading else "{0:b}".format(value)

    def _fromDEC(self, value: int, leading: bool = False):
        """
        returns a String representation of a Decimal Int including the corresponding
        Hexadecimal and Binary number.
        """
        return str(value) + " {Hexadecimal: " + str(self.__dec_to_hex__(value, leading)) + "; Binary: " + str(self.__dec_to_bin__(value, leading)) + "}"

    def __hex_to_dec__(self, value: str):
        return int(value, 16)

    def __hex_to_bin__(self, value: str, leading:bool = False):
        return bin(int(value, 16)) if leading else bin(int(value, 16))[2:]

    def _fromHEX(self, value: str, leading: bool = False):
        """
        returns a String representation of a Hexadecimal String including the corresponding
        Decimal and Binary number.
        """
        return value + " {Decimal: " + str(self.__hex_to_dec__(value)) + "; Binary: " + str(self.__hex_to_bin__(value, leading)) + "}"

    def __bin_to_dec__(self, value: str):
        return int(value, 2)

    def __bin_to_hex__(self, value: str, leading: bool = False):
        return hex(int(value, 2)) if leading else hex(int(value, 2))[2:]

    def _fromBIN(self, value: str, leading: bool = False):
        """
        returns a String representation of a Binary String including the corresponding
        Decimal and Hexadecimal number.
        """
        return str(value) + " {Decimal: " + str(self.__bin_to_dec__(value)) + "; Hexadecimal: " + str(self.__bin_to_hex__(value, leading)) + "}"


import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README_PyPI.md")) as readme:
    README = readme.read()

setup(
    name='cat_win',
    version='1.0.4',    
    description="Simple 'cat' Command-line Tool made in Python",
    url='https://github.com/SilenZcience/cat_win',
    author='Silas A. Kraume',
    author_email='silas.kraume1552@gmail.com',
    long_description=README,
    long_description_content_type='text/markdown',
    license='MIT License',
    packages=['cat_win', 'cat_win.util'],
    install_requires=['pyperclip3==0.4.1'],
    entry_points={
        'console_scripts': 'cat = cat_win.cat:main'
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6'
)